/*
 * KPC488 driver implementation
 * Copyright (C) 1994  W. Eric Norum
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 * 
 * W. Eric Norum
 * Saskatchewan Accelerator Laboratory
 * University of Saskatchewan
 * 107 North Road
 * Saskatoon, Saskatchewan, CANADA
 * S7N 5C6
 * 
 * eric@skatter.usask.ca
 */
 
#import "KPC488.h"

#define DDM_DEBUG	1

#import <kernserv/prototypes.h>
#import <sys/errno.h>

#import <driverkit/generalFuncs.h>
#import	<driverkit/interruptMsg.h>
#import	<driverkit/kernelDriver.h>
#import <driverkit/i386/ioPorts.h>
#import <driverkit/i386/kernelDriver.h>
#import <driverkit/debugging.h>

#import "upd7210.h"
#import "IEEE488Messages.h"
#import "KPC488ioctl.h"

/*
 * I/O thread commands
 */
enum ioCommands {
	CMD_OPEN,
	CMD_CLOSE,
	CMD_READ,
	CMD_WRITE,
	CMD_IOCTL,
	CMD_SPOLL,
};

/*
 * Local DMA buffer
 */
#define DMA_BUFSIZE		(8*1024)	/* Pagesize? IOMallocLow contiguous? */

/*
 * Parameter keys (as listed in Default.table)
 */
#define CONTROLLER_KEY	"Controller GPIB Address"
#define TIMEOUT_KEY		"Time Limit"
#define DMACHECK_KEY	"DMA Check Interval"

/*
 * Dynamic Debugging Module
 */
#if (defined (DDM_DEBUG))
# define DDM_INDEX	0
# define DDM_PROGRESS	0x100
# define DDM_COMMANDS	0x200
# define DDM_INTERRUPTS	0x400

#define logProgress(x,a,b,c)	IODEBUG(DDM_INDEX,DDM_PROGRESS,x,a,b,c,0,0)
#define logCommands(a)			IODEBUG(DDM_INDEX,DDM_COMMANDS,"CMD:0x%x\n",a,0,0,0,0)
#define logInterrupts(x,a,b,c,d)	IODEBUG(DDM_INDEX,DDM_INTERRUPTS,x,a,b,c,d,0)
#else
#define logProgress(fmt,a,b,c)
#define logCommands(a)
#define logInterrupts(x,a,b,c,d)
#endif

/*
 * Hardware port access
 */
#define UPDIN(p)	inb(isaPorts.start+(p))
#define UPDOUT(p,v)	outb(isaPorts.start+(p),(v))

/*
 * Interthread communication (commandLock conditions)
 */
#define COMMAND_IDLE	0
#define COMMAND_IO		1
#define COMMAND_DONE	2

/*
 * Assume only one device
 */
static KPC488 *sc;

/*
 * UNIX-style entry points
 */
static int KPC488open (dev_t dev, int flag, int mode);
static int KPC488close (dev_t dev, int flag, int mode);
static int KPC488read (dev_t dev, struct uio *uio, int flag);
static int KPC488write (dev_t dev, struct uio *uio, int flag);
static int KPC488select (dev_t dev, int rw);
static int KPC488ioctl (dev_t dev, int cmd, caddr_t data, int flag);

@implementation KPC488

/*
 * Set up uPD7210 hardware
 */
- (void)init7210
{
	/*
	 * Reset chip
	 */
	UPDOUT (UPDOUT_AUX, UPD_AUX_CR);

	/*
	 * Set timeout values (on-board 8 MHz oscillator)
	 */
	UPDOUT (UPDOUT_AUX, UPD_AUX_FREQ(8));

	/*
	 * Allow high-speed transfers
	 */
	UPDOUT (UPDOUT_AUX, UPD_AUX_B(0x4));

	/*
	 * Set interface function to T/L, normal pullup enable
	 */
	UPDOUT (UPDOUT_AMR, UPD_AMR_TRM1 | UPD_AMR_TRM0 | UPD_AMR_ADM0);
	UPDOUT (UPDOUT_ADR01, myAddress);
	UPDOUT (UPDOUT_ADR01, UPD_ADR01_ARS | UPD_ADR01_DT | UPD_ADR01_DL);

}

/*
 ************************************************************************
 *							DriverKit methods							*
 ************************************************************************
 */

/*
 * Probe, configure board and init new instance.
 */
+ (BOOL)probe:description
{
	KPC488 *			dev;
	BOOL				ret;

	/*
	 * Allocate an instance
	 */
	dev = [self alloc];
	if (dev == nil) {
		IOLog ("%s: Failed to allocate instance\n", [self name]) ;
		return NO;
	}

	/*
	 * Initialize the instance
	 */
	if ([dev initFromDeviceDescription:description] == nil) {
		IOLog ("%s: Failed to init Device Description\n", [self name]);
		[dev free];
		return NO;
	}

	/*
	 * Link the instance into the kernel
	 */
	ret = [self addToCdevswFromDescription:description
										open:  (IOSwitchFunc)KPC488open	
										close: (IOSwitchFunc)KPC488close
										read:  (IOSwitchFunc)KPC488read
										write: (IOSwitchFunc)KPC488write
										ioctl: (IOSwitchFunc)KPC488ioctl
										stop:  (IOSwitchFunc)nulldev
										reset: (IOSwitchFunc)nulldev
										select:(IOSwitchFunc)KPC488select
										mmap:  (IOSwitchFunc)nodev
										getc:  (IOSwitchFunc)nodev
										putc:  (IOSwitchFunc)nodev];
	if (ret != YES) {
		IOLog ("%s: addToCdevswFromDescription failed.\n", [self name]);
		[dev free];
		return NO;
    }
	IOLog ("%s: addToCdevswFromDescription major %d\n", [self name], [self characterMajor]);
	return YES;
}

- (BOOL)getInt:(unsigned int *)ip fromConfig:config stringKey:(const char *)key
{
	const char *value, *cp;

	if ((value = [config valueForStringKey:key]) == NULL) {
		IOLog ("%s: No value for key `%s'\n", [self name], key);
		return NO;
	}
	*ip = 0;
	for (cp = value ; *cp != '\0' ; cp++) {
		if ((*cp >= '0') && (*cp <= '9')) {
			*ip = *ip * 10 + *cp - '0';
		}
		else {
			IOLog ("%s: Bad value (%s) for key `%s'\n", [self name], value, key);
			[config freeString:value];
			return NO;
		}
	}
	[config freeString:value];
	return YES;
}

/*
 * Initialize the new instance.
 */
- initFromDeviceDescription:deviceDescription
{
	IOConfigTable *config = [deviceDescription configTable];
	IOReturn r;
	unsigned int i;
	char cbuf[40];

    if ([super initFromDeviceDescription:deviceDescription] == nil)
    	return nil;

	/*
	 * Verify that the description is reasonable
	 */
	if (([deviceDescription numInterrupts] != 1)
	 || ([deviceDescription numChannels] != 1)
	 || ([deviceDescription numMemoryRanges] != 0)
	 || ([deviceDescription numPortRanges] != 1)) {
		IOLog ("%s: Bad device description!\n", [[self class] name]);
		return nil;
	}
	isaPorts = *[deviceDescription portRangeList];
	if (isaPorts.size != 8) {
		IOLog ("%s: Bad number of ports in device description!\n", [[self class] name]);
		return nil;
	}
	if ([deviceDescription interrupt] == 2) {
		int	aList[1];
		aList[0] = 9;
		[deviceDescription setInterruptList:aList num:1];
	}

    [self setName:[config valueForStringKey:"Driver Name"]];
    [self setUnit:0];
	sprintf (cbuf, "%x-%x", isaPorts.start, isaPorts.start+isaPorts.size-1);
    [self setLocation:cbuf];
    IOLog ("%s: Interrupt:%d Channel:%d\n",	[self name],
											[deviceDescription interrupt], 
											[deviceDescription channel]);

	if (![self getInt:&myAddress fromConfig:config stringKey:CONTROLLER_KEY])
		return nil;
	if (![self getInt:&timeLimit fromConfig:config stringKey:TIMEOUT_KEY])
		return nil;
	if (![self getInt:&i fromConfig:config stringKey:DMACHECK_KEY])
		return nil;
	IOLog ("%s: Controller:%d  Timeout:%d sec  DMA Check:%d msec\n",
												[self name], myAddress,
												timeLimit, i);
	dmaCheckInterval = i;
	dmaCheckInterval *= 1000*1000;

	/*
	 * Start the I/O thread
	 */
    r = [self startIOThread];
    if (r != IO_R_SUCCESS)
		IOLog ("%s: Can't start I/O thread: %s\n", [self name], [self stringFromReturn:r]);

	/*
	 * Set up inter-thread communication
	 */
	commandLock = [[NXConditionLock alloc] initWith:COMMAND_IDLE];
	commandMessage.msg_id = IO_COMMAND_MSG;
	commandMessage.msg_size = sizeof commandMessage;
	commandMessage.msg_type = MSG_TYPE_NORMAL;
	commandMessage.msg_remote_port = IOConvertPort ([self interruptPort],
													IO_KernelIOTask, IO_Kernel);
	commandMessage.msg_local_port = PORT_NULL;

	timeoutMessage.msg_id = IO_TIMEOUT_MSG;
	timeoutMessage.msg_size = sizeof timeoutMessage;
	timeoutMessage.msg_type = MSG_TYPE_NORMAL;
	timeoutMessage.msg_remote_port = IOConvertPort ([self interruptPort],
													IO_KernelIOTask, IO_Kernel);
	timeoutMessage.msg_local_port = PORT_NULL;

	dmaInterruptMessage.msg_id = IO_DEVICE_INTERRUPT_MSG;
	dmaInterruptMessage.msg_size = sizeof dmaInterruptMessage;
	dmaInterruptMessage.msg_type = MSG_TYPE_NORMAL;
	dmaInterruptMessage.msg_remote_port = IOConvertPort ([self interruptPort],
													IO_KernelIOTask, IO_Kernel);
	dmaInterruptMessage.msg_local_port = PORT_NULL;

	sc = self;

	/*
	 * Set up hardware resources
	 */
	if ([self isEISAPresent]) {
		r = [self setDMATransferWidth:IO_8Bit forChannel:0];
		if (r != IO_R_SUCCESS) {
			IOLog ("%s: Can't set DMA transfer width:%s\n", [self name], [self stringFromReturn:r]);
			return nil;
		}
		r = [self setDMATiming:IO_TypeA forChannel:0];
		if (r != IO_R_SUCCESS) {
			IOLog ("%s: Can't set DMA timing :%s\n", [self name], [self stringFromReturn:r]);
			return nil;
		}
	}
	r = [self setTransferMode:IO_Single forChannel:0];
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't set DMA transfer mode:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
	dmaBuffer = IOMallocLow (DMA_BUFSIZE);
	if (dmaBuffer == NULL) {
		IOLog ("Can't IOMallocLow DMA buffer\n");
		return nil;
	}
	r = IOPhysicalFromVirtual (IOVmTaskSelf(), (vm_address_t)dmaBuffer, &dmaBufferPhysicalAddress);
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't get DMA buffer physical address:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
    [self registerDevice];

	/*
	 * Set up the hardware
	 */
	[self init7210];
	r = [self enableInterrupt:0];
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't enable interrupt:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
	r = [self enableChannel:0];
	if (r != IO_R_SUCCESS)  {
		IOLog ("%s: Can't enable channel:%s\n", [self name], [self stringFromReturn:r]);
		return nil;
	}
    return self; 
}

/*
 * Pass values to user-level routines
 */
- (IOReturn)getIntValues:(unsigned int *)values
				forParameter:(IOParameterName)parameter
				count:(unsigned int *)count
{
	if ((strcmp (parameter, "IOMajorDevice") == 0) && (*count == 1)) {
		values[0] = [[self class] characterMajor];
		return IO_R_SUCCESS;
	}
	else if ((strcmp (parameter, CONTROLLER_KEY) == 0) && (*count == 1)) {
		values[0] = myAddress;
		return IO_R_SUCCESS;
	}
	else if ((strcmp (parameter, TIMEOUT_KEY) == 0) && (*count == 1)) {
		values[0] = timeLimit;
		return IO_R_SUCCESS;
	}
	else if ((strcmp (parameter, DMACHECK_KEY) == 0) && (*count == 1)) {
		values[0] = dmaCheckInterval / (1000 * 1000);
		return IO_R_SUCCESS;
	}
	return [super getIntValues:values forParameter:parameter count:count];
}

/*
 * Get values from user-level routines
 */
- (IOReturn)setIntValues:(unsigned int *)values
				forParameter:(IOParameterName)parameter
				count:(unsigned int)count
{
	if ((strcmp (parameter, TIMEOUT_KEY) == 0) && (count == 1)) {
		if ((values[0] >= 1) && (values[0] <= 100)) {
			timeLimit = values[0];
			return IO_R_SUCCESS;
		}
		else {
			return IO_R_INVALID_ARG;
		}
	}
	else if ((strcmp (parameter, DMACHECK_KEY) == 0) && (count == 1)) {
		if ((values[0] >= 1) && (values[0] <= 100)) {
			dmaCheckInterval = values[0];
			dmaCheckInterval *= (1000 * 1000);
			return IO_R_SUCCESS;
		}
		else {
			return IO_R_INVALID_ARG;
		}
	}
	return [super setIntValues:values forParameter:parameter count:count];
}

/*
 ************************************************************************
 *								Timer Functions							*
 ************************************************************************
 */
static void
KPC488CheckDMA (void *arg)
{
	struct KPC488 *self = arg;
	IOReturn r;

	logProgress ("CheckDMA state:%d\n", self->ioState, 0, 0);
	self->checkingDMA = NO;
	/*
	 * It would seem to make more sense to call isDMADone: here and
	 * only send a message when the DMA is complete, but after some helpful
	 * information from a very kind person at NeXT I've determined that
	 * doing so is a bad idea.
	 * isDMADone is broken and currentCountForChannel is unsafe to call.
	 */
	if (self->ioState == IOS_DATA) {
		r = msg_send_from_kernel (&self->dmaInterruptMessage, MSG_OPTION_NONE, 0);
		if (r != SEND_SUCCESS)
			IOLog ("Can't send DMA done message to I/O thread: %d\n", r);
	}
}

static void
KPC488Timeout (void *arg)
{
	struct KPC488 *self = arg;
	IOReturn r;

	if (self->ioState == IOS_IDLE)
		return;
	if (self->checkingDMA) {
		self->checkingDMA = NO;
		ns_untimeout ((func)KPC488CheckDMA, self);
	}
	r = msg_send_from_kernel (&self->timeoutMessage, MSG_OPTION_NONE, 0);
	if (r != SEND_SUCCESS)
		IOLog ("Can't send timeout message to I/O thread: %d\n", r);
}

/*
 ************************************************************************
 *							I/O Thread Methods							*
 ************************************************************************
 */
- (void)finishClosing
{
	logProgress ("Cease GPIB operation\n", 0, 0, 0);
	UPDOUT (UPDOUT_IMR2, 0);			/* Disable interrupts */
	UPDOUT (UPDOUT_IMR1, 0);
	UPDOUT (UPDOUT_AUX, UPD_AUX_CREN);	/* Clear remote enable */
}

- (void)interfaceClear
{
	logProgress ("IFC\n", 0, 0, 0);
	lastTalker = lastListener = -1;
	UPDOUT (UPDOUT_AUX, UPD_AUX_SIFC);	/* Set IFC */
	IODelay (100);						/* Wait for 100 usec */
	UPDOUT (UPDOUT_AUX, UPD_AUX_CIFC);	/* Clear IFC */
}

- (void)finishSPOLL
{
	logProgress ("End SPOLL: FHS\n", 0, 0, 0);
	UPDOUT (UPDOUT_IMR1, UPD_IMR1_END);	/* Restore interrupt enables */
	UPDOUT (UPDOUT_IMR2, UPD_IMR2_SRQI|UPD_IMR2_DMAO|UPD_IMR2_DMAI|UPD_IMR2_CO);
	UPDOUT (UPDOUT_AUX, UPD_AUX_FHS);	/* Finish handshake */
	UPDOUT (UPDOUT_AUX, UPD_AUX_A (0));	/* Disable RFD holdoff */
}

- (void)reportCompletionWithError:(int)errno
{
	if (errno) {
		if (ioEISADmaBuffer) {
			logProgress ("Terminate with DMA count:%d\n", [self currentCountForChannel:0], 0, 0
);
			[self abortDMABuffer:ioEISADmaBuffer];
			ioEISADmaBuffer = NULL;
		}
		if (checkingDMA) {
			self->checkingDMA = NO;
			ns_untimeout ((func)KPC488CheckDMA, self);
		}
		if (ioState == IOS_SPOLL)
			[self finishSPOLL];
		[self interfaceClear];
		logProgress ("LUN TCA\n", 0, 0, 0);
		UPDOUT (UPDOUT_AUX, UPD_AUX_LUN);
		UPDOUT (UPDOUT_AUX, UPD_AUX_TCA);
	}
	ioState = IOS_IDLE;
	commandErrno = errno;
	if (openFlags == 0)
		[self finishClosing];
	[commandLock unlockWith:COMMAND_DONE];
}

- (void)completeCommand:(int)errno
{
	IOUnscheduleFunc (KPC488Timeout, self);
	[self reportCompletionWithError:errno];
}

- (void)timeoutOccurred
{
	/*
	 * If command has finished, just return.
	 */
	if (ioState == IOS_IDLE)
		return;

	/*
	 * Report the timeout
	 */
	logProgress ("Timeout: ADS:0x%x\n", UPDIN (UPDIN_ADS), 0, 0);
	[self reportCompletionWithError:ETIMEDOUT];
}

- (void)sendNextCommandByte
{
	softCOflag = NO;
	if (ieee488CommandCount) {
		logCommands (ieee488Commands[ieee488CommandIndex]);
		UPDOUT (UPDOUT_BO, ieee488Commands[ieee488CommandIndex++]);
	}
	else {
		logProgress ("EPP\n", 0, 0, 0);
		ioState = IOS_PPOLL;
		UPDOUT (UPDOUT_AUX, UPD_AUX_EPP);
	}
}

/*
 * Send a sequence of commands.  Operation depends on value of first argument:
 *	1) count > 0
 *		Subsequent `count' arguments are the multiline messages
 *		to be placed in ieee488Commands and transmitted.
 *	2) count == 0
 *		Conduct a parallel poll.
 *	3) count < 0
 *		ieee488Commands and ieee488CommandCount are already set up.
 */
- (void)sendCommands:(int)count, ...
{
	va_list ap;

	va_start (ap, cmd);
	if (UPDIN (UPDIN_ADS) & UPD_ADS_ATNBAR) {
		logProgress ("TCA\n", 0, 0, 0);
		UPDOUT (UPDOUT_AUX, UPD_AUX_TCA);
	}
	ioState = IOS_COMMAND;
	ieee488CommandIndex = 0;
	if (count > 0) {
		ieee488CommandCount = 0;
		while (count--)
			ieee488Commands[ieee488CommandCount++] = va_arg (ap, int);
	}
	else if (count == 0) {
		ieee488CommandCount = 0;
	}
	if (softCOflag == YES)
		[self sendNextCommandByte];
}

- (int)startDMA
{
	IOReturn r;
	unsigned int phys = dmaBufferPhysicalAddress;
	int ads;
	BOOL EOIflag = NO;

	ioState = IOS_DATA;
	if (uio->uio_resid <= DMA_BUFSIZE) {
		dmaLength = uio->uio_resid;
		if ((commandType == CMD_WRITE) && (uio->uio_resid > 1))
			dmaLength--;	/* Have to set EOI before sending last byte */
	}
	else {
		dmaLength = DMA_BUFSIZE;
	}
	ioEISADmaBuffer = [self createDMABufferFor:&phys
											length:dmaLength
											read:commandType == CMD_READ
											needsLowMemory:YES
											limitSize:YES];
	logProgress ("Phys:0x%x  phys:0x%x length:%d\n", dmaBufferPhysicalAddress, phys, dmaLength);
	if (ioEISADmaBuffer == NULL)
		return ENOSPC;
	ads = UPDIN (UPDIN_ADS);
	if ((commandType == CMD_WRITE) && (uio->uio_resid == 1)) {
		logProgress ("SEOI\n", 0, 0, 0);
		UPDOUT (UPDOUT_AUX, UPD_AUX_SEOI);
		EOIflag = YES;
	}
	r = [self startDMAForBuffer:ioEISADmaBuffer channel:0];
    if (r != IO_R_SUCCESS) {
		[self abortDMABuffer:ioEISADmaBuffer];
		IOLog ("%s: Can't start DMA: %s\n", [self name], [self stringFromReturn:r]);
		ioEISADmaBuffer = NULL;
		return EIO;
	}
	if ((ads & UPD_ADS_ATNBAR) == 0) {
		if (commandType == CMD_READ) {
			logProgress ("ADS:0x%x TCSE GTS\n", ads, 0, 0);
			UPDOUT (UPDOUT_AUX, UPD_AUX_TCSE);
		}
		else {
			logProgress ("ADS:0x%x TCSE LTNC GTS\n", ads, 0, 0);
			UPDOUT (UPDOUT_AUX, UPD_AUX_TCSE);
			UPDOUT (UPDOUT_AUX, UPD_AUX_LTNC);
		}
		UPDOUT (UPDOUT_AUX, UPD_AUX_GTS);
		softCOflag = NO;
	}
	if (!EOIflag) {
		checkingDMA = YES;
		ns_timeout ((func)KPC488CheckDMA, self, dmaCheckInterval, CALLOUT_PRI_SOFTINT0);
	}
	return 0;
}

- (void)interruptOccurred
{
	int isr2;
	int isr1;
	unsigned int count;
	int err;

	/*
	 * Must read ISR2 first
	 */
	isr2 = UPDIN (UPDIN_ISR2);
	isr1 = UPDIN (UPDIN_ISR1);
	logInterrupts ("ISR1:0x%x ISR2:0x%x ADS:0x%x State:%d\n",
														isr1, isr2,
														UPDIN (UPDIN_ADS),
														ioState);
	if (isr2 & UPD_ISR2_SRQI)
		softSRQIflag = YES;
	if (isr2 & UPD_ISR2_CO) {
		switch (ioState) {
		case IOS_COMMAND:
			if (ieee488CommandCount && (ieee488CommandIndex == ieee488CommandCount)) {
				softCOflag = YES;
				switch (commandType) {
				case CMD_SPOLL:
					if (ieee488Commands[0] != SPD_488) {
						/*
						 * Switch from DMA to Interrupt-driven input
						 */
						UPDOUT (UPDOUT_IMR1, UPD_IMR1_END|UPD_IMR1_DI);
						UPDOUT (UPDOUT_IMR2, UPD_IMR2_SRQI|UPD_IMR2_CO);
						UPDOUT (UPDOUT_AUX, UPD_AUX_A (1));	/* Enable RFD holdoff */
						logProgress ("LTN GTS\n", 0, 0, 0);
						UPDOUT (UPDOUT_AUX, UPD_AUX_LTN);
						UPDOUT (UPDOUT_AUX, UPD_AUX_GTS);
						softCOflag = NO;
						ioState = IOS_SPOLL;
					}
					else {
						[self completeCommand:0];
					}
					break;

				case CMD_IOCTL:
					[self completeCommand:0];
					break;

				case CMD_CLOSE:
					[self completeCommand:0];
					break;

				case CMD_READ:
				case CMD_WRITE:
					if ((err = [self startDMA]) != 0)
						[self completeCommand:err];
					break;
				}
			}
			else {
				[self sendNextCommandByte];
			}
			break;

		case IOS_PPOLL:
			softCOflag = YES;
			*(int *)ioctlData = UPDIN (UPDIN_CPT);
			logProgress ("PPOLL:0x%x\n", *(int *)ioctlData, 0, 0);
			[self completeCommand:0];
			break;

		default:
			softCOflag = YES;
			break;
		}
	}
	switch (ioState) {
	case IOS_IDLE:
	case IOS_COMMAND:
	case IOS_PPOLL:
		break;

	case IOS_SPOLL:
		if (isr1 & UPD_ISR1_DI) {
			*(int *)ioctlData = UPDIN (UPDIN_DI);
			logProgress ("SPOLL:0x%x TCS\n", *(int *)ioctlData, 0, 0);
			UPDOUT (UPDOUT_AUX, UPD_AUX_TCS);
			[self finishSPOLL];
			[self sendCommands:1, SPD_488];
		}
		break;

	case IOS_DATA:
		count = [self currentCountForChannel:0];
		if ((isr1 & UPD_ISR1_END) || (count == 0)) {
			unsigned int ndone = dmaLength - count;
			err = 0;
			if (self->checkingDMA) {
				self->checkingDMA = NO;
				ns_untimeout ((func)KPC488CheckDMA, self);
			}
			uio->uio_resid -= ndone;
			logProgress ("Free DMA after %d bytes, %d left, (0x%x).\n",
											ndone, uio->uio_resid,
											[self currentAddressForChannel:0]);
			[self freeDMABuffer:ioEISADmaBuffer];
			ioEISADmaBuffer = NULL;
			if (isr1 & UPD_ISR1_ERR) {
				err = EIO;
			}
			else if (isr1 & UPD_ISR1_END) {
				lastTalker = lastListener = -1;
				if (commandType == CMD_WRITE) {
					logProgress ("FHS\n", 0, 0, 0);
					UPDOUT (UPDOUT_AUX, UPD_AUX_FHS);
				}
				else {
					readFinished |= (1 << target);
				}
			}
			else {
				if (commandType == CMD_WRITE) {
					if (uio->uio_resid == 1) {
						dmaBuffer[0] = dmaBuffer[ndone];
						logProgress ("Copy last character (0x%x)\n", dmaBuffer[0], 0, 0);
						if ((err = [self startDMA]) == 0)
							break;
					}
				}
			}
			[self completeCommand:err];
		}
		else {
			if (self->checkingDMA == NO) {
				self->checkingDMA = YES;
				ns_timeout ((func)KPC488CheckDMA, self, dmaCheckInterval, CALLOUT_PRI_SOFTINT0);
			}
		}
		break;
	}
}

- (int)startRead
{
	int err;

	IOScheduleFunc (KPC488Timeout, self, timeLimit);
	readFinished &= ~(1 << target);
	if ((lastTalker != target) || (lastListener != myAddress)) {
		lastTalker = target;
		lastListener = myAddress;
		logProgress ("Send commands for read.\n", 0, 0, 0);
		[self sendCommands:4, UNT_488,
								UNL_488,
								TAG_488 + target,
								LAG_488 + myAddress];
	}
	else {
		if ((err = [self startDMA]) != 0)
			return err;
	}
	return 0;
}

- (int)startWrite
{
	int err;

	IOScheduleFunc (KPC488Timeout, self, timeLimit);
	if ((lastTalker != myAddress) || (lastListener != target)) {
		lastTalker = myAddress;
		lastListener = target;
		logProgress ("Send commands for write.\n", 0, 0, 0);
		[self sendCommands:4, UNT_488,
								UNL_488,
								TAG_488 + myAddress,
								LAG_488 + target];
	}
	else {
		if ((err = [self startDMA]) != 0)
			return err;
	}
	return 0;
}

- (void)makeListenerList:(int)listeners
{
	int bit, i;

	bit = 0x1;
	i = 0;
	while ((listeners != 0) && (i <= 31)) {
		if (listeners & bit) {
			listeners &= ~bit;
			if (i != myAddress)
				ieee488Commands[ieee488CommandCount++] = LAG_488 + i;
		}
		i++;
		bit <<= 1;
	}
}

- (void)startIEEE488Operation
{
	lastTalker = lastListener = -1;
	IOScheduleFunc (KPC488Timeout, self, timeLimit);
	switch (ioctlCmd) {
	case KPC_IEEE488_GTL:
		[self sendCommands:3, UNL_488, LAG_488+target, GTL_488];
		break;

	case KPC_IEEE488_SDC:
		[self sendCommands:3, UNL_488, LAG_488+target, SDC_488];
		break;

	case KPC_IEEE488_GET:
		[self sendCommands:3, UNL_488, LAG_488+target, GET_488];
		break;

	case KPC_IEEE488_LLO:
		[self sendCommands:1, LLO_488];
		break;

	case KPC_IEEE488_DCL:
		[self sendCommands:1, DCL_488];
		break;

	case KPC_IEEE488_SPOLL:
		commandType = CMD_SPOLL;
		[self sendCommands:3, UNL_488, SPE_488, TAG_488+target];
		break;

	case KPC_IEEE488_PPU:
		[self sendCommands:1, PPU_488];
		break;

	case KPC_IEEE488_PPD:
		[self sendCommands:3, UNL_488, LAG_488+target, PPD_488];
		break;

	case KPC_IEEE488_PPC:
		[self sendCommands:5, UNL_488, LAG_488+target, PPC_488,
									PPE_488+(*(int *)ioctlData & 0xF), UNL_488];
		break;

	case KPC_IEEE488_PPOLL:
		[self sendCommands:0];
		break;

	case KPC_IEEE488_GETLIST:
		ieee488CommandCount = 0;
		ieee488Commands[ieee488CommandCount++] = UNL_488;
		[self makeListenerList:*(int *)ioctlData];
		ieee488Commands[ieee488CommandCount++] = GET_488;
		ieee488Commands[ieee488CommandCount++] = UNL_488;
		[self sendCommands:-1];
		break;
	}
}

- (int)startIoctl
{
	switch (ioctlCmd) {
	case KPC_IEEE488_GTL:
	case KPC_IEEE488_SDC:
	case KPC_IEEE488_GET:
	case KPC_IEEE488_LLO:
	case KPC_IEEE488_DCL:
	case KPC_IEEE488_SPOLL:
	case KPC_IEEE488_PPU:
	case KPC_IEEE488_PPD:
	case KPC_IEEE488_PPC:
	case KPC_IEEE488_PPOLL:
	case KPC_IEEE488_GETLIST:
		[self startIEEE488Operation];
		break;

	default:
		return EINVAL;
	}
	return 0;
}

- (void)forceLocal
{
	logProgress ("Force LOCAL\n", 0, 0, 0);
	lastTalker = lastListener = -1;
	UPDOUT (UPDOUT_AUX, UPD_AUX_CREN);	/* Clear REN */
	IODelay (100);						/* Wait for 100 usec */
	UPDOUT (UPDOUT_AUX, UPD_AUX_SREN);	/* Set REN */
}

- (void)commandRequestOccurred
{
	int err = 0;

	[commandLock lockWhen:COMMAND_IO];
	switch (commandType) {
	case CMD_OPEN:
		if (openFlags == 0) {
			logProgress ("Initialize With ports 0x%x (%d)\n", isaPorts.start, isaPorts.size, 0);
			softCOflag = NO;
			softSRQIflag = NO;
			ioState = IOS_IDLE;
			UPDOUT (UPDOUT_AUX, UPD_AUX_PON);	/* Set chip to power-on state */
			[self interfaceClear];
			UPDOUT (UPDOUT_IMR1, UPD_IMR1_END);	/* Enable interrupts */
			UPDOUT (UPDOUT_IMR2, UPD_IMR2_SRQI|UPD_IMR2_DMAO|UPD_IMR2_DMAI|UPD_IMR2_CO);
			UPDOUT (UPDOUT_AUX, UPD_AUX_SREN);	/* Set remote enable */
		}
		openFlags |= 1 << target;
		break;

	case CMD_CLOSE:
		openFlags &= ~(1 << target);
		lastTalker = lastListener = -1;
		IOScheduleFunc (KPC488Timeout, self, timeLimit);
		[self sendCommands:4, UNL_488,
									TAG_488 + myAddress,
									LAG_488 + target,
									GTL_488];
		return;

	case CMD_READ:
		if ((err = [self startRead]) == 0)
			return;
		break;

	case CMD_WRITE:
		if ((err = [self startWrite]) == 0)
			return;
		break;

	case CMD_IOCTL:
		switch (ioctlCmd) {
		case KPC_IEEE488_IFC:
			[self interfaceClear];
			break;
		
		case KPC_IEEE488_LOCAL:
			[self forceLocal];
			break;
		
		case KPC_IEEE488_GETSRQ:
			*(int *)ioctlData = softSRQIflag;
			softSRQIflag = NO;
			break;
		
		default:
			if ((err = [self startIoctl]) == 0)
				return;
			break;
		}
		break;

	default:
		err = EINVAL;
		break;
	}
	[self completeCommand:err];
}

/*
 ************************************************************************
 *	Interface between UNIX-style entry points and DriverKit methods		*
 ************************************************************************
 */
- (int)KPC488ExecuteCommand:(char)cmd forDevice:(int)device
{
	kern_return_t r;

	/*
	 * Set up information common to all commands
	 */
	target = device;
	commandType = cmd;

	/*
	 * Tell the I/O thread about the new command
	 */
	[commandLock unlockWith:COMMAND_IO];
	r = msg_send_from_kernel (&commandMessage, MSG_OPTION_NONE, 0);
	if (r != SEND_SUCCESS) {
		IOLog ("Can't send message to I/O thread: %d\n", r);
		return EIO;
	} 

	/*
	 * Await completion
	 */
	[commandLock lockWhen:COMMAND_DONE];

	/*
	 * Return the results
	 */
	return commandErrno;
}

@end

/*
 ************************************************************************
 *			Good old-fashioned UNIX driver entry points					*
 ************************************************************************
 */
static int
KPC488open (dev_t dev, int flag, int mode)
{	
	int err;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	err = [sc KPC488ExecuteCommand:CMD_OPEN forDevice:minor(dev)];
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
KPC488close (dev_t dev, int flag, int mode)
{
	int err;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	err = [sc KPC488ExecuteCommand:CMD_CLOSE forDevice:minor(dev)];
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
KPC488read (dev_t dev, struct uio *uio, int flag)
{
	struct iovec *iov = uio->uio_iov;
	int ndone, nleft, ncopy;
	int err = 0;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	sc->uio = uio;
	nleft = ndone = 0;
	while (uio->uio_resid) {
		if (nleft == 0) {
			nleft = uio->uio_resid;
			err = [sc KPC488ExecuteCommand:CMD_READ forDevice:minor(dev)];
			if (err)
				break;
			ndone = 0;
			nleft -= uio->uio_resid;
		}
		if (iov->iov_len > nleft)
			ncopy = nleft;
		else
			ncopy = iov->iov_len;
		if (copyout (sc->dmaBuffer + ndone, iov->iov_base, ncopy)) {
			err = EFAULT;
			break;
		}
		ndone += ncopy;
		nleft -= ncopy;
		iov->iov_base += ncopy;
		iov->iov_len -= ncopy;
		if (iov->iov_len == 0) {
			uio->uio_iovcnt--;
			iov++;
		}
		if (sc->readFinished & (1 << minor (dev)))
			break;
	}
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
KPC488write (dev_t dev, struct uio *uio, int flag)
{
	struct iovec *iov = uio->uio_iov;
	int ndone, nleft, ncopy;
	int err = 0;

	[sc->commandLock lockWhen:COMMAND_IDLE];
	sc->uio = uio;
	nleft = DMA_BUFSIZE;
	ndone = 0;
	while (uio->uio_resid) {
		if (iov->iov_len > nleft)
			ncopy = nleft;
		else
			ncopy = iov->iov_len;
		if (copyin (iov->iov_base, sc->dmaBuffer + ndone, ncopy)) {
			err = EFAULT;
			break;
		}
		ndone += ncopy;
		nleft -= ncopy;
		iov->iov_base += ncopy;
		iov->iov_len -= ncopy;
		if (iov->iov_len == 0) {
			uio->uio_iovcnt--;
			iov++;
		}
		if ((nleft == 0) || (uio->uio_iovcnt == 0)) {
			err = [sc KPC488ExecuteCommand:CMD_WRITE forDevice:minor(dev)];
			if (err)
				break;
			ndone = 0;
			nleft = DMA_BUFSIZE;
		}
	}
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}

static int
KPC488select (dev_t dev, int rw)
{
	return EIO;
}

static int
KPC488ioctl (dev_t dev, int cmd, caddr_t data, int flag)
{
	int err;

	if (cmd == KPC_IEEE488_END) {
		*(int *)data = ((sc->readFinished & (1 << minor (dev))) != 0);
		return 0;
	}
	[sc->commandLock lockWhen:COMMAND_IDLE];
	sc->ioctlCmd = cmd;
	sc->ioctlData = data;
	err = [sc KPC488ExecuteCommand:CMD_IOCTL forDevice:minor(dev)];
	[sc->commandLock unlockWith:COMMAND_IDLE];
	return err;
}
