/*
 * Name: Par.m
 * Description: loadable kernel-server for handling of parallel port
 *     (replacement for NeXT's ParallelPort)
 * Author: Christian Starkjohann <cs@ds1.kph.tuwien.ac.at>
 * Date: September 1 1994
 * Tabsize: 4
 *
 * Copyright (C) 1994  Christian Starkjohann
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * Changelog:
 * Version 1.0:
 * 1994-09-05
 *   changed interrupt-priority to IPLTTY (=IPL1)
 * 1994-09-08
 *   introduced task_sleeps-variable to avoid ns_abstimeout()-table-overflow
 *   (Timeouts are called from a 20ms-timer-interrupt, and the table is
 *   about 100 entries big. In the rare case that the output-fifo-charge
 *   is constantly around the wakeup-threshold, more than 100 wakeups could
 *   be scheduled within 20ms and the system crashes.)
 * Version 1.1:
 * 1994-09-28
 *   corrected FIFO_CHARGE()-macro so that it always returns the correct
 *   value
 * 1994-10-06
 *   PPIOCSETTIMO-ioctl only supported on demand, because many drivers set
 *   this to stupid 2 seconds, enabling debug is easier now.
 *   Some fine-tuning of the README-file.
 * Version 1.2:
 * 1994-11-23
 *   modified TASK_WAKEUP()-macro to do wakeup only if task is really sleeping
 *   (don't like that, but maybe it helps against some kinds of crashes...)
 * 1995-01-10
 *   changed the strobe-pulse from 1 to 2 microseconds
 * 1995-01-24
 *   Changed IPL form IPLTTY and spltty() to IPL=2 and spl2(), because in the
 *   3.3-kernel IPLTTY and spltty() don't match any more.
 *   Inserted mutual exclusion code in interrupt that also gives an
 *   error-message if spln() does not work.
 *   Changed timeout-handling because long long divide is not supported for
 *   kernel-servers.
 *   Took TASK_SLEEP() out of else-branch in par_write(). This optimization
 *   reduces the call-frequency for copyin().
 * Version 1.3:
 * 1996-01-14
 *   Finally found the race condition that may cause a system panic: changed
 *   FIFO_WRITE_CHUNK() macro.
 *   Added supervision of interrupt line.
 * Version 1.4:
 * 1996-02-16, Axel Habermann
 *   Added support for I/O hardware with broken interrupts. Done some 
 *   performance improvements. Set version number to 1.4 (Axel Habermann,
 *   kiwi@buran.fb10.tu-berlin.de)
 * 1996-02-19, Christian Starkjohann:
 *   Converted Axel's changes to my code conventions, solved 2 problems:
 *     1. ENABLE was called without calling DISABLE before (removed all
 *        locking, ENABLE and DISABLE are not needed any more!)
 *     2. Some printers need considerable time from receiving the first byte
 *        to an ACK for this byte (measured 3.5 seconds on a HP Deskjet+).
 *        During this time, the printer is _not_ busy. This triggered the
 *        'broken interrupt' detector. Changed the detection code and increased
 *        ACK timeout to about 6 seconds.
 *   Changed logging mechanism.
 * 1996-03-14, Christian Starkjohann:
 *   Set version number to 1.5, after several weeks of beta test.
 */

#import "Par.h"
#import <driverkit/generalFuncs.h>
#import <driverkit/devsw.h>
#import <kernserv/printf.h>
#import <kernserv/i386/spl.h>
#import <kernserv/lock.h>
#import <driverkit/i386/ioPorts.h>

#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/syslog.h>
#include <kernserv/prototypes.h>
#include <kernserv/sched_prim.h>
#import <dev/i386/pp_extern.h>

#import "../config.h"
#import "hardware.h"

/* ------------------------------------------------------------------------- */
/* ---------------------- configuration constants -------------------------- */
/* ------------------------------------------------------------------------- */
#define INSTANCE_CNT    8       /* maximum number of instances */
#define TIMER_PERIOD    200     /* supervisory timer-interrupt-period in ms */
#define SHORT_TIMER     20      /* period in ms when running without irqs */
#define PP_TIMEOUT      180     /* default-timeout in seconds */
#define RECHECK         50      /* spin that long in check status (max) */

#define BUFFER_SIZE     4096    /* size of output-buffer in bytes */
#define WAKEUP_MARK     1000    /* low-watermark to wakeup writing task */

#define MY_IPL          2       /* interrupt-priority for this interrupt */
#define DISABLE         ipl = spl2()    /* interrupt-disable-function */
#define ENABLE          splx(ipl)       /* interrupt-enable-function */

#ifdef DEBUG
#define DLOG(p)         IOLog p
#else
#define DLOG(p)
#endif

/* ------------------------------------------------------------------------- */
/* ----------------------------- structures -------------------------------- */
/* ------------------------------------------------------------------------- */

struct fifo{                /* data-structure for fifo-style ringbuffers */
    char    *buffer;        /* pointer to allocated buffer */
    int     rp;             /* index of read-position in the buffer */
    int     wp;             /* index of write-position in the buffer */
    int     size;           /* size of the allocated buffer */
};

/* ------------------------------------------------------------------------- */

struct instance_data{       /* static data for each instance */
    int         iobase;     /* base of the io-address-space */
    int         irq;        /* number of the interrupt used */
    struct fifo fifo;       /* fifo-buffer for output-data */
    int         octl;       /* status of writable hardware-bits */
    int         is_open;    /* flag, true if device is opened */
    int         timeout;    /* timeout to use in seconds */
    int         is_dead;    /* flag, true if printer has timeout */
    int         task_sleeps;/* flag, true if task is sleeping */
    int         missing_int;/* # of printing par_start()s since last intr. */
    int         recheck;    /* time to wait in interrupt for printer ready */
    int         bytes_hardint;  /* number of bytes written from hard-intr. */
    int         bytes_softint;  /* number of bytes written from timer-intr. */
    int         real_irqs;  /* number of IRQs during the job */
    int         irq_latency;/* number of timer intr til first hard intr */
};

/* ------------------------------------------------------------------------- */
/* ----------------------------- global data ------------------------------- */
/* ------------------------------------------------------------------------- */

static char DeviceName[] = DEVICENAME;
static char DeviceKind[] = "ParallelPort";

static volatile struct instance_data    instance_data[INSTANCE_CNT] = {0};
static int                              instances = 0;
static int                              ParMajor;

/* ------------------------------------------------------------------------- */
/* ------------------------ function prototypes ---------------------------- */
/* ------------------------------------------------------------------------- */

static int  par_open(dev_t dev, int flag, int mode);
static int  par_close(dev_t dev, int flag, int mode);
static int  par_read(dev_t dev, struct uio *uio, int flag);
static int  par_write(dev_t dev, struct uio *uio, int flag);
static int  par_ioctl(dev_t dev, int cmd, caddr_t data, int flag);
static void timer_interrupt(int dummy);
static void par_interrupt(int dummy);
static void par_start(int dev, int is_softint);
extern int  intr_register_irq(int Irq, IOEISAInterruptHandler *Handler,
                                                dev_t Unit, unsigned int Ipl);
extern int  intr_unregister_irq(int Irq);
extern int  intr_enable_irq(int Irq);
extern int  intr_disable_irq(int Irq);

/* ------------------------------------------------------------------------- */
/* ----------------------- FIFO-handling-Macros ---------------------------- */
/* ------------------------------------------------------------------------- */

#define FIFO_NEXT_P(fifo, p)    (((fifo).p + 1) % (fifo).size)
#define FIFO_CLEAR(fifo)        ((fifo).rp = (fifo).wp = 0)
#define FIFO_WRITABLE(fifo)     (FIFO_NEXT_P(fifo, wp) != (fifo).rp)
#define FIFO_CHARGE(fifo)       (((fifo).wp - (fifo).rp + (fifo).size)  \
                                                            % (fifo).size)
#define FIFO_READABLE(fifo)     ((fifo).wp != (fifo).rp)
#define FIFO_CHUNK_WRITTEN(fifo, len)   ((fifo).wp=((fifo).wp+len)%(fifo).size)

#define FIFO_GET(fifo, c)                                       \
        do{ if(FIFO_READABLE(fifo)){                            \
                c = (fifo).buffer[(fifo).rp];                   \
                (fifo).rp = FIFO_NEXT_P(fifo, rp);              \
            }else                                               \
                c = 0;                                          \
        }while(0)

#define FIFO_WRITE_CHUNK(fifo, len, base)                       \
        do{ int d = (fifo).rp - (fifo).wp;                      \
            if(d <= 0)                                          \
                len = (fifo).size-(fifo).wp-((fifo).rp==0?1:0); \
            else                                                \
                len = d - 1;                                    \
            base = (fifo).buffer + (fifo).wp;                   \
        }while(0)


/* ------------------------------------------------------------------------- */
/* --------------------------- function-macros ----------------------------- */
/* ------------------------------------------------------------------------- */

#define TASK_SLEEP(dev)                                                     \
        do{                                                                 \
            assert_wait((int)&(instance_data[dev].fifo), YES);              \
            instance_data[dev].task_sleeps = YES;                           \
            thread_block();                                                 \
        }while(0)

#define TASK_WAKEUP(interrupt, dev)                                         \
        do{                                                                 \
            if(instance_data[dev].task_sleeps){                             \
                if(interrupt)                                               \
                    ns_timeout((func)thread_wakeup,                         \
                                (void *)&(instance_data[dev].fifo), 1,      \
                                        CALLOUT_PRI_SOFTINT0);              \
                else                                                        \
                    thread_wakeup((int)&(instance_data[dev].fifo));         \
                instance_data[dev].task_sleeps = NO;                        \
            }                                                               \
        }while(0)

/* ------------------------------------------------------------------------- */
/* ------------------------- objective-C-interface ------------------------- */
/* ------------------------------------------------------------------------- */

@implementation Par

/* ------------------------------------------------------------------------- */

+ (BOOL)probe:description       /* called by the driver-loader */
{
Par             *dev;
IODirectDevice  *desc;

    if(instances >= INSTANCE_CNT){
        IOLog("%s%d: too many instances\n", DeviceName, instances);
        return NO;
    }
    dev = [self alloc];
    if(dev == nil){
        IOLog("%s%d: Failed to allocate an instance\n", DeviceName, instances);
        return NO;
    }
    [dev setInstance:instances];
    desc = [dev initFromDeviceDescription:description];
    if(desc == nil){
        IOLog("%s%d: Failed to init Device\n", DeviceName, instances);
        [dev free];
        return NO;
    }
    instance_data[instances].fifo.buffer = kalloc(BUFFER_SIZE);
    if(instance_data[instances].fifo.buffer == NULL){
        IOLog("%s%d: Failed to allocate buffer\n", DeviceName, instances);
        [dev free];
        return NO;
    }
    instance_data[instances].fifo.size = BUFFER_SIZE;
    instance_data[instances].timeout = PP_TIMEOUT;
    FIFO_CLEAR(instance_data[instances].fifo);

    instances++;    /* this instance was created successfully */
    return YES;
}

/* ------------------------------------------------------------------------- */

- initFromDeviceDescription:description
{
int     iobase, iosize, irq;
IORange *pr1;
char    my_name[sizeof(DeviceName) + 8];
id      desc;

/*
 * parse the Instance?-table for relevant data:
 * io-base and length, interrupt number
 */
    if([description numPortRanges] < 1){
        IOLog("%s%d: no Port-range in description\n", DeviceName, instance);
        return nil;
    }
    pr1 = [description portRangeList];
    iobase = pr1->start;
    iosize = pr1->size;
    if(iosize < 3){
        IOLog("%s%d: port-size too small in description\n", DeviceName,
                                                                    instance);
        return nil;
    }
    if([description numInterrupts] < 1){
        IOLog("%s%d: no Interrupts in description\n", DeviceName, instance);
        return nil;
    }
    irq = [description interrupt];
    sprintf(my_name, "%s%1d", DeviceName, instance);
/*
 * Initialize some data for the superclass
 */
    [self setName:my_name];
    [self setDeviceKind:DeviceKind];
    [self setUnit:instance];
    [self setLocation:NULL];

    instance_data[instance].octl = PP_NOTRESET;
    outb(PORT_OCTL, instance_data[instance].octl);
/*
 * make switchtable-entrys
 */
    if (instance == 0) {
        ParMajor = IOAddToCdevsw(
            (IOSwitchFunc)par_open,         /* IOSwitchFunc openFunc         */
            (IOSwitchFunc)par_close,        /* IOSwitchFunc closeFunc        */
            (IOSwitchFunc)par_read,         /* IOSwitchFunc readFunc         */
            (IOSwitchFunc)par_write,        /* IOSwitchFunc writeFunc        */
            (IOSwitchFunc)par_ioctl,        /* IOSwitchFunc ioctlFunc        */
            nodev,                          /* IOSwitchFunc stopFunc         */
            nulldev,                        /* IOSwitchFunc resetFunc        */
            nulldev,                        /* IOSwitchFunc selectFunc       */
            nodev,                          /* IOSwitchFunc mmapFunc         */
            nodev,                          /* IOSwitchFunc getcFunc         */
            nodev);                         /* IOSwitchFunc putcFunc         */
        if (ParMajor == -1) {
            IOLog("%s%d failed to place into Cdevsw\n", DeviceName, instance);
            return nil;
        }
    }
/*
 * register the device
 */
    [self registerDevice];

    desc = [super initFromDeviceDescription:description];
    if(desc == nil){
        IOLog("%s%d superclass init failed\n", DeviceName, instance);
        return nil;
    }
/*
 * register our interrupt-routine
 */
    intr_unregister_irq(irq);
    if(!intr_register_irq(irq, (IOEISAInterruptHandler *) par_interrupt,
                                                    instance, MY_IPL)){
        IOLog("%s%d: Failed to register IRQ %d\n", DeviceName, instance, irq);
        return nil;
    }
    if(!intr_enable_irq(irq)){
        IOLog("%s%d: Failed to enable IRQ %d\n", DeviceName, instance, irq);
        return nil;
    }
    instance_data[instance].iobase = iobase;
    instance_data[instance].irq = irq;
    DLOG(("%s%d at iobase 0x%x IRQ %d, major %d\n", DeviceName, instance,
                                                    iobase, irq, ParMajor));
/*
 * start the timer-interrupt
 */
    if (instance == 0){
        ns_timeout((func)timer_interrupt, 0, (TIMER_PERIOD)*1000*1000,
                                                        CALLOUT_PRI_SOFTINT0);
    }
    return desc;
}

/* ------------------------------------------------------------------------- */

- (void)setInstance:(int)inst   /* needed to access instance-variable */
{
    instance = inst;
}

/* ------------------------------------------------------------------------- */

/*
 * This method is an interface to the Post-Loader. The Post-Loader can ask
 * for integer-arrays by name. The only name supported is 'GeneralInfo':
 * it returns the major-devicenumber, our instance-number and the number of
 * instances.
 */
- (IOReturn) getIntValues:(unsigned int *)values
    forParameter:(IOParameterName) parameter
    count:(unsigned int *)count
{
int     i, maxCount = *count;

    if (maxCount == 0) {
        maxCount = IO_MAX_PARAMETER_ARRAY_LENGTH;
    }
    if (strcmp(parameter, "GeneralInfo") == 0) {
        i = 0;
        if(i < maxCount) values[i++] = ParMajor;
        if(i < maxCount) values[i++] = instance;
        if(i < maxCount) values[i++] = instances;
        *count = i;
        return IO_R_SUCCESS;
    }
    return[super getIntValues:values
           forParameter:parameter
           count:&maxCount];
}

/* ------------------------------------------------------------------------- */

/*
 * Get system-time in seconds (nanoseconds are a little unhandy...).
 * Use approximation: shift ns-time by 30 bits. This is necessary because
 * long long division is not supported for the kernel. And this with all the
 * long long declared kernel-routines.... :-(
 */
static int  systime_s(void)
{
long long   tm;

    tm = (long long)clock_value(System);
    tm = tm >> 30;
    return (int) tm;
}


/* ------------------------------------------------------------------------- */
/* ------------------------ switchtable-routines --------------------------- */
/* ------------------------------------------------------------------------- */


static int  par_open(dev_t dev, int flag, int mode)
{
int iobase;

    dev = minor(dev);
    if(dev < 0 || dev >= instances){
        DLOG(("%s%d: par_open: dev out of range\n", DeviceName, dev));
        return ENXIO;
    }
    iobase = instance_data[dev].iobase;
    if(instance_data[dev].is_open){     /* only exclusive use permitted */
        DLOG(("%s%d: par_open: already open\n", DeviceName, dev));
        return EBUSY;
    }
    FIFO_CLEAR(instance_data[dev].fifo);
    instance_data[dev].octl |= PP_IEN ;             /* enable interrupt */
    outb(PORT_OCTL, instance_data[dev].octl);
    instance_data[instances].timeout = PP_TIMEOUT;  /* reset timeout */
    instance_data[dev].is_dead = NO;
    instance_data[dev].bytes_hardint = 0;           /* reset counters */
    instance_data[dev].bytes_softint = 0;
    instance_data[dev].real_irqs = 0;
    instance_data[dev].irq_latency = 0;
    instance_data[dev].recheck = RECHECK;
    instance_data[dev].is_open = YES;
    DLOG(("%s%d: par_open: success\n", DeviceName, dev));
    return 0;
}

/* ------------------------------------------------------------------------- */

static int  par_close(dev_t dev, int flag, int mode)
{
int     iobase;
int     timeout_tm;

    dev = minor(dev);
    if(dev < 0 || dev >= instances){
        DLOG(("%s%d: par_close: dev out of range\n", DeviceName, dev));
        return ENXIO;
    }
    iobase = instance_data[dev].iobase;
    timeout_tm = systime_s() + instance_data[dev].timeout;
/*
 * flush output-data before closing
 */
    while(FIFO_READABLE(instance_data[dev].fifo)
                                            && !instance_data[dev].is_dead){
        if(instance_data[dev].timeout > 0 && (timeout_tm - systime_s()) < 0){
            DLOG(("%s%d: par_close: timeout\n", DeviceName, dev));
            break;
        }
        TASK_SLEEP(dev);
    }
    instance_data[dev].octl &= ~PP_IEN;         /* disable interrupt */
    outb(PORT_OCTL, instance_data[dev].octl);
    instance_data[dev].is_open = NO;
    IOLog("%s%d: finished job: %d rechecks, %d bytes (%d from timer), "
        "%d irqs, irq latency: %d\n", DeviceName, dev,
        instance_data[dev].recheck, instance_data[dev].bytes_hardint
        + instance_data[dev].bytes_softint, instance_data[dev].bytes_softint,
        instance_data[dev].real_irqs, instance_data[dev].irq_latency);
    DLOG(("%s%d: par_close: success\n", DeviceName, dev));
    return 0;
}

/* ------------------------------------------------------------------------- */

static int  par_read(dev_t dev, struct uio *uio, int flag)
{
    dev = minor(dev);
    DLOG(("%s%d: par_read: read not supported!\n", DeviceName, dev));
    return EINVAL;  /* no read from this device possible */
}

/* ------------------------------------------------------------------------- */

static int  par_write(dev_t dev, struct uio *uio, int flag)
{
long    userlen, len;
char    *userbase, *dest;
int     timeout_tm;

    dev = minor(dev);
    if(dev < 0 || dev >= instances){
        DLOG(("%s%d: par_write: dev out of range\n", DeviceName, dev));
        return ENXIO;
    }
    if(uio->uio_iovcnt != 1){       /* single requests only */
        DLOG(("%s%d: par_write: multiple blocks\n", DeviceName, dev));
        log(LOG_INFO, "%s%d: uio->uio_iovcnt = %d occured\n", DeviceName, dev,
                                                            uio->uio_iovcnt);
        return EINVAL;
    }
    if(uio->uio_iov->iov_len == 0){
        DLOG(("%s%d: par_write: NULL write\n", DeviceName, dev));
        return 0;           /* nothing to do */
    }

    userbase = uio->uio_iov->iov_base;
    userlen = uio->uio_iov->iov_len;
    timeout_tm = systime_s() + instance_data[dev].timeout;
    while(!instance_data[dev].is_dead){
        if(FIFO_WRITABLE(instance_data[dev].fifo)){
            timeout_tm = systime_s() + instance_data[dev].timeout;
            FIFO_WRITE_CHUNK(instance_data[dev].fifo, len, dest);
            if(len == 0){
                DLOG(("%s%d: par_write: error: len = 0\n", DeviceName, dev));
                break;
            }
            if(len > userlen)
                len = userlen;
            copyin(userbase, dest, len);
            FIFO_CHUNK_WRITTEN(instance_data[dev].fifo, len);
            par_start(dev, YES);
            userlen -= len;
            userbase += len;
        }
        if(userlen <= 0)
            break;
        TASK_SLEEP(dev);
        if(instance_data[dev].timeout > 0 && (timeout_tm - systime_s()) < 0){
            DLOG(("%s%d: par_write: timeout\n", DeviceName, dev));
            instance_data[dev].is_dead = YES;
            break;
        }
    }
    uio->uio_resid = userlen;   /* number of bytes that could not be written */
    if(userlen == 0){
        DLOG(("%s%d: par_write: success\n", DeviceName, dev));
        return 0;
    }
    else{
        DLOG(("%s%d: par_write: failed\n", DeviceName, dev));
        return EDEVERR;         /* timeout occured */
    }
}

/* ------------------------------------------------------------------------- */

static int  par_ioctl(dev_t dev, int cmd, caddr_t data, int flag)
{
int     *arg = (int *) data;
int     iobase, hw, mask;

    dev = minor(dev);
    if(dev < 0 || dev >= instances){
        DLOG(("%s%d: par_ioctl: dev out of range\n", DeviceName, dev));
        return ENXIO;
    }
    iobase = instance_data[dev].iobase;
    switch(cmd){
    case PPIOCSW:       /* read status word (int)*/
        DLOG(("%s%d: par_ioctl: get status\n", DeviceName, dev));
        hw = inb(PORT_ICTL);
        mask = PP_ST_INITIALIZED;
        if(!(hw & PP_NOTBUSY))
            mask |= PP_ST_BUSY;
        if(hw & PP_NOPAPER)
            mask |= PP_ST_NOPAPER;
        if(!(hw & PP_ONLINE))
            mask |= PP_ST_NOTSELECTED;
        if(!(hw & PP_NOTERROR))
            mask |= PP_ST_ERROR;
        *arg = mask;
        break;
    case PPIOCINIT:     /* initialize printer */
        DLOG(("%s%d: par_ioctl: reset\n", DeviceName, dev));
        outb(PORT_OCTL, instance_data[dev].octl & ~PP_NOTRESET);
        DELAY(20);
        outb(PORT_OCTL, instance_data[dev].octl);
        break;
    case PPIOCSETTIMO:  /* set timeout period (secs) (int)*/
        DLOG(("%s%d: par_ioctl: set timeout to %ds\n", DeviceName, dev, *arg));
#ifdef ALLOW_SET_TIMEOUT
        instance_data[dev].timeout = *arg;
#endif
        break;
    default:
        DLOG(("%s%d: par_ioctl: unknown ioctl\n", DeviceName, dev));
        return EINVAL;
    }
    DLOG(("%s%d: par_ioctl: success\n", DeviceName, dev));
    return 0;
}

/* ------------------------------------------------------------------------- */

static void timer_interrupt(int dummy)
{
int     i;
char    no_irq = NO;
    
    for(i=0;i<instances;i++){
        par_start(i, YES);
        no_irq |= instance_data[i].missing_int < 0;
        TASK_WAKEUP(NO, i);
    }
    ns_timeout((func)timer_interrupt, 0, (no_irq ? SHORT_TIMER : TIMER_PERIOD) 
                                        *1000*1000, CALLOUT_PRI_SOFTINT0);
}

/* ------------------------------------------------------------------------- */

static void par_interrupt(int dummy)    /* called for each interrupt */
{
int     i;

    for(i=0;i<instances;i++){
        if(instance_data[i].missing_int > 0){
            if(instance_data[i].missing_int > instance_data[i].irq_latency)
                instance_data[i].irq_latency = instance_data[i].missing_int;
            instance_data[i].missing_int = 0;
        }
        par_start(i, NO);
    }
}

/* ------------------------------------------------------------------------- */

/*
 * par_start() does not need to change the interrupt priority for the
 * following reasons:
 * - for direct calls: an interrupt before in_sensitive_area was set to YES
 *   does not matter. If it occures after in_sensitive_area was set to YES,
 *   the interrupt routine terminates immediately.
 * - for timer interrupts: same as before is valid for hardware interrupts,
 *   direct calls can not interrupt timer interrupts. If the interrupt routine
 *   finds out that it interrupts a direct call, it terminates.
 * - for hardware interrupts: same as for timer interrupts. Hardware interrupts
 *   can not be interrupted by timer interrupts or direct calls.
 * Why not write several bytes in one hardware interrupt:
 * - We would have to disable the hardware interrupt during this time,
 *   otherwise the still occuring interrupts would cancel out the efficiency
 *   improvement. But disabling the interrupt is not safe. We would possibly
 *   lose interrutps.
 * - If the printer is slow, waiting within the interrupt does not help.
 *   It would reduce efficiency.
 * - Interrupt routines should be as short as possiblem, because all lower
 *   priority interrupts are disabled during that time.
 */

static void par_start(int dev, int softint)
{
int                     i, c, iobase = instance_data[dev].iobase;
char                    inc_missingint = YES;
static volatile char    in_sensitive_area[INSTANCE_CNT] = {0};

    if(!instance_data[dev].is_open)     /* device must be open */
        return;
    if(!softint)
        instance_data[dev].real_irqs++;
    if(in_sensitive_area[dev]){
        DLOG(("%s%d: par_start: sensitive area hit by %s interrupt!\n",
                        DeviceName, dev, softint ? "timer" : "hardware"));
        return;
    }
    in_sensitive_area[dev] = YES;

    /* write i bytes at once */
    i = softint ? 1022 : 10;    /* try only a few when running at IPL2 */
    while(i-- > 0 && (inb(PORT_ICTL) & PP_NOTBUSY)){
        if(inc_missingint && instance_data[dev].missing_int >= 0
                                    && instance_data[dev].missing_int++ > 30){
            instance_data[dev].missing_int = -1;
            IOLog("%s%d: IRQ %d (or printer\'s ACK line) does not work!\n",
                                    DeviceName, dev, instance_data[dev].irq);
        }
        inc_missingint = NO;
        instance_data[dev].is_dead = NO;        /* printer was not busy */
        if(FIFO_READABLE(instance_data[dev].fifo)){
            FIFO_GET(instance_data[dev].fifo, c);
            outb(PORT_DATA, c);
            if(softint)
                instance_data[dev].bytes_softint++;
            else
                instance_data[dev].bytes_hardint++;
            outb(PORT_OCTL, instance_data[dev].octl | PP_STROBE);
            DELAY(2);                   /* strobe-pulse of this length  */
            outb(PORT_OCTL, instance_data[dev].octl);
            if(FIFO_CHARGE(instance_data[dev].fifo) == WAKEUP_MARK){
                /* wakeup task to refill the fifo */
                TASK_WAKEUP(YES, dev);
            }
            /* only when running at IPL0 we try to convince the printer
               to take some more bytes ;-) by asking several times in a row */
            if(softint){
                int     dc = instance_data[dev].recheck;
                while(!(inb(PORT_ICTL) & PP_NOTBUSY) && dc > 0) dc--;
                if(dc > 3 && instance_data[dev].recheck > 3)
                    instance_data[dev].recheck--;
                else if(instance_data[dev].recheck < RECHECK)
                    instance_data[dev].recheck += 3;
            }
        }else
            break;      /* no more data in the fifo */
    }
    in_sensitive_area[dev] = NO;
}

/* ------------------------------------------------------------------------- */

@end

